@@ -81,11 +81,11 @@ static pa_sample_spec sample_spec = {
     .rate = 44100,
     .channels = 2
 };
-static const pa_buffer_attr custom_bufattr ={
+static pa_buffer_attr custom_bufattr ={
     .maxlength = (uint32_t)-1,
     .minreq = (uint32_t)-1,
     .prebuf = (uint32_t)-1,
-    .fragsize = 4096,
+    .fragsize = 2048,
     .tlength = 4096
 };
 const pa_buffer_attr * bufattr = NULL;
@@ -100,6 +100,11 @@ static void playback_stream_drain_cb(pa_stream *s, int success, void *userdata);
 
 void pacat_log(const char *fmt, ...) {
     va_list args;
+    struct timespec ts;
+
+    clock_gettime(CLOCK_REALTIME, &ts);
+
+    fprintf(stderr, "%10jd.%06ld ", ts.tv_sec, ts.tv_nsec / 1000);
     va_start(args, fmt);
     vfprintf(stderr, fmt, args);
     va_end(args);
@@ -210,6 +215,16 @@ static void process_playback_data(struct userdata *u, pa_stream *s, size_t max_l
         return;
     }
 
+    if (verbose > 1) {
+        pa_usec_t latency = 0;
+        int negative;
+
+        if (pa_stream_get_latency(s, &latency, &negative))
+            pacat_log("pa_stream_get_latency() failed");
+
+        pacat_log("process_playback_data(): vchan data %d max_length %d latency %llu", space_in_vchan, max_length, latency);
+    }
+
     buffer_length = (size_t)space_in_vchan > max_length ? max_length : (size_t)space_in_vchan;
     if (!buffer_length)
         goto maybe_cork;
@@ -664,6 +679,8 @@ static void context_state_callback(pa_context *c, void *userdata) {
             pa_stream_set_event_callback(u->play_stream, stream_event_callback, u);
             pa_stream_set_buffer_attr_callback(u->play_stream, stream_buffer_attr_callback, u);
             flags = PA_STREAM_ADJUST_LATENCY;
+            if (verbose > 1)
+                flags |= PA_STREAM_AUTO_TIMING_UPDATE | PA_STREAM_INTERPOLATE_TIMING;
 
             if (pa_stream_connect_playback(u->play_stream, u->play_device, bufattr, flags, NULL /* volume */, NULL) < 0) {
                 pacat_log("pa_stream_connect_playback() failed: %s", pa_strerror(pa_context_errno(c)));
@@ -1008,21 +1025,26 @@ static void vchan_rec_async_connect(pa_mainloop_api *UNUSED(a),
     u->mainloop_api->io_free(u->rec_ctrl_event);
     u->rec_ctrl_event = NULL;
 
-    if (u->never_block == -1)
-        u->never_block = libvchan_buffer_space(u->rec_ctrl) > 8192;
-
     /* when both vchans are connected, connect to the daemon */
     if (libvchan_is_open(u->play_ctrl) == VCHAN_CONNECTED)
         connect_pa_daemon(u);
 }
 
+void cleanup_pidfile(char *pidfile_path, int pidfile_fd) {
+    unlink(pidfile_path);
+    close(pidfile_fd);
+}
+
 static _Noreturn void usage(char *arg0, int arg) {
     FILE *stream = arg ? stderr : stdout;
-    fprintf(stream, "usage: %s [-l] [--] domid domname\n",
+    fprintf(stream, "usage: %s [options] [--] domid domname\n",
             arg0 ? arg0 : "pacat-simple-vchan");
     fprintf(stream, "  -l - low-latency mode (higher CPU usage)\n");
-    fprintf(stream, "  -n - never block on vchan I/O (default if record buffer size is 8192 bytes or more)\n");
-    fprintf(stream, "  -b - always block on vchan I/O (default if record buffer size is less than 8192 bytes)\n");
+    fprintf(stream, "  -n - never block on vchan I/O (overrides previous -b option)\n");
+    fprintf(stream, "  -b - always block on vchan I/O (default, overrides previous -n option)\n");
+    fprintf(stream, "  -v - verbose logging (a lot of output, may affect performance)\n");
+    fprintf(stream, "  -t size - target playback buffer fill, implies -l, default %d\n",
+            custom_bufattr.tlength);
     fprintf(stream, "  -h - print this message\n");
     if (fflush(NULL) || ferror(stdout) || ferror(stderr))
         exit(1);
@@ -1035,26 +1057,40 @@ int main(int argc, char *argv[])
     struct userdata u;
     pa_glib_mainloop* m = NULL;
     pa_time_event *time_event = NULL;
-    int domid = -1;
     char *pidfile_path;
     int pidfile_fd;
-    int play_watch_fd, rec_watch_fd;
     int i;
+    unsigned long tlength;
+    char *endptr;
 
     memset(&u, 0, sizeof(u));
-    u.never_block = -1;
     if (argc <= 2)
         usage(argv[0], 1);
-    while ((i = getopt(argc, argv, "+lnbh")) != -1) {
+    while ((i = getopt(argc, argv, "+lnbvt:h")) != -1) {
         switch (i) {
             case 'l':
                 bufattr = &custom_bufattr;
                 break;
             case 'n':
-                u.never_block = 1;
+                u.never_block = true;
                 break;
             case 'b':
-                u.never_block = 0;
+                u.never_block = false;
+                break;
+            case 'v':
+                verbose += 1;
+                break;
+            case 't':
+                errno = 0;
+                tlength = strtoul(optarg, &endptr, 0);
+                if (*endptr)
+                    errx(1, "Invalid -t argument: %s", optarg);
+                if (tlength > UINT32_MAX)
+                    errno = ERANGE;
+                if (errno)
+                    err(1, "Invalid -t argument: %s", optarg);
+                bufattr = &custom_bufattr;
+                custom_bufattr.tlength = tlength;
                 break;
             case 'h':
                 usage(argv[0], 0);
@@ -1065,37 +1101,41 @@ int main(int argc, char *argv[])
     if (argc - optind != 2)
         usage(argv[0], 1);
     const char *domid_str = argv[optind], *domname = argv[optind + 1];
-    char *endptr;
     errno = 0;
     long l_domid = strtol(domid_str, &endptr, 10);
     /* 0x7FF0 is DOMID_FIRST_RESERVED */
     if (l_domid < 0 || l_domid >= 0x7FF0 || errno == ERANGE)
         errx(1, "domid %ld out of range 0 through %d inclusive", l_domid,
                 0x7FF0 - 1);
-    domid = l_domid;
     if (errno)
         err(1, "invalid domid %s", domid_str);
     if (*endptr)
         errx(1, "trailing junk after domid %s", domid_str);
-    if (create_pidfile(domid, &pidfile_path, &pidfile_fd) < 0)
+
+    u.domid = (int)l_domid;
+    u.name = domname;
+
+    if (create_pidfile(u.domid, &pidfile_path, &pidfile_fd) < 0)
         /* error already printed by create_pidfile() */
         exit(1);
+    goto main;
 
+main:
     u.ret = 1;
 
     g_mutex_init(&u.prop_mutex);
 
-    u.name = domname;
-
-    u.play_ctrl = libvchan_client_init_async(domid, QUBES_PA_SINK_VCHAN_PORT, &play_watch_fd);
+    u.play_ctrl = libvchan_client_init_async(u.domid, QUBES_PA_SINK_VCHAN_PORT, &u.play_watch_fd);
     if (!u.play_ctrl) {
         perror("libvchan_client_init_async");
-        exit(1);
+        cleanup_pidfile(pidfile_path, pidfile_fd);
+        exit(u.ret);
     }
-    u.rec_ctrl = libvchan_client_init_async(domid, QUBES_PA_SOURCE_VCHAN_PORT, &rec_watch_fd);
+    u.rec_ctrl = libvchan_client_init_async(u.domid, QUBES_PA_SOURCE_VCHAN_PORT, &u.rec_watch_fd);
     if (!u.rec_ctrl) {
         perror("libvchan_client_init_async");
-        exit(1);
+        cleanup_pidfile(pidfile_path, pidfile_fd);
+        exit(u.ret);
     }
     if (setgid(getgid()) < 0) {
         perror("setgid");
@@ -1129,15 +1169,15 @@ int main(int argc, char *argv[])
         goto quit;
     }
 
-    u.play_ctrl_event = u.mainloop_api->io_new(u.mainloop_api,
-            play_watch_fd, PA_IO_EVENT_INPUT, vchan_play_async_connect, &u);
+    u.play_ctrl_event = u.mainloop_api->io_new(
+        u.mainloop_api, u.play_watch_fd, PA_IO_EVENT_INPUT, vchan_play_async_connect, &u);
     if (!u.play_ctrl_event) {
         pacat_log("io_new play_ctrl failed");
         goto quit;
     }
 
-    u.rec_ctrl_event = u.mainloop_api->io_new(u.mainloop_api,
-            rec_watch_fd, PA_IO_EVENT_INPUT, vchan_rec_async_connect, &u);
+    u.rec_ctrl_event = u.mainloop_api->io_new(
+        u.mainloop_api, u.rec_watch_fd, PA_IO_EVENT_INPUT, vchan_rec_async_connect, &u);
     if (!u.rec_ctrl_event) {
         pacat_log("io_new rec_ctrl failed");
         goto quit;
@@ -1167,14 +1207,21 @@ quit:
         control_cleanup(&u);
     }
 
-    if (u.play_stream)
+    if (u.play_stream) {
         pa_stream_unref(u.play_stream);
+        u.play_stream = NULL;
+    }
 
-    if (u.rec_stream)
+    if (u.rec_stream) {
         pa_stream_unref(u.rec_stream);
+        u.rec_stream = NULL;
+    }
 
-    if (u.context)
+    if (u.context) {
+        pa_context_disconnect(u.context);
         pa_context_unref(u.context);
+        u.context = NULL;
+    }
 
     if (time_event) {
         assert(u.mainloop_api);
@@ -1219,7 +1266,10 @@ quit:
 
     g_mutex_clear(&u.prop_mutex);
 
-    unlink(pidfile_path);
-    close(pidfile_fd);
+    if (!u.ret)
+        goto main;
+
+    cleanup_pidfile(pidfile_path, pidfile_fd);
+
     return u.ret;
 }
